# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Utilities for using backends."""
from ._basebackend import BaseBackend
from ._backendutils import (get_backend_class,
                            get_backend_instance,
                            configuration,
                            calibration,
                            parameters,
                            status,
                            local_backends,
                            remote_backends,
                            register_backend,
                            discover_local_backends,
                            discover_remote_backends,
                            update_backends)
